#!/usr/bin/env python

from distutils.core import setup

setup(
    name = 'gearman',
    version = '1.0',
    description = 'Gearman Client',
    author = 'Samuel Stauffer',
    author_email = 'samuel@descolada.com',
    url = 'http://danga.com/gearman/',
    packages = ['gearman'],
)
